/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "stdafx.h"

#ifdef _USRDLL
#define _QUBFITS_LIB_
#endif 

#include "qubfits.h"
#include "QUB_BasicCurve.h"
#include "QUB_Curve_Expo.h"
#include "QUB_Curve_DecExpo.h"
#include "QUB_Curve_Gaussian.h"
#include "QUB_LMFitter.h"

extern "C" QUBFITS_API int    __stdcall initqubfits()
{
	RegisterCurve("Linear", CreateLinear);
	RegisterCurve("Exponential", CreateExpo);
	RegisterCurve("Exponential-Linear", CreateExpoLinear);
	RegisterCurve("Declining Exponential", CreateDecExpo);
	RegisterCurve("Double Declining Exponential", CreateDblDecExpo);
	RegisterCurve("Gaussian", CreateGaussian);
	RegisterCurve("Gaussian-Linear", CreateGaussianLinear);

	RegisterFitter("Levenburg-Marquardt", CreateLMFitter);

	return 0;
}


IQUB_Evts::~IQUB_Evts()
{}

typedef list<IQUB_EvtListener*> ListenList;
typedef map<string, ListenList> Listeners;

QUB_Evts::QUB_Evts()
{
	listeners = new Listeners;
}

QUB_Evts::~QUB_Evts()
{
	delete ((Listeners*) listeners);
}

void __cdecl QUB_Evts::Free()
{
	delete this;
}

void __cdecl QUB_Evts::add(const char *name, IQUB_EvtListener *rcvr)
{
	Listeners& listeners = * ((Listeners*) this->listeners);
	listeners[name].push_back(rcvr);
}

void __cdecl QUB_Evts::del(const char *name, IQUB_EvtListener *rcvr)
{
	Listeners& listeners = * ((Listeners*) this->listeners);
	listeners[name].remove(rcvr);
}

void __cdecl QUB_Evts::send(const char *name, void *sender)
{
	Listeners& listeners = * ((Listeners*) this->listeners);
	ListenList& ll = listeners[name];
	for (ListenList::iterator li = ll.begin(); li != ll.end(); ++li)
		(*li)->QUBEvt(name, sender);
}

extern "C" QUBFITS_API IQUB_Evts *    __stdcall QUB_Evts_Create()
{
	return new QUB_Evts;
}


map<string, QUB_CurveFactory> g_curveMakers;
map<string, QUB_FitterFactory> g_fitterMakers;

extern "C" QUBFITS_API void __stdcall RegisterCurve(const char *name, QUB_CurveFactory factory)
{
	g_curveMakers[name] = factory;
}

extern "C" QUBFITS_API IQUB_Curve* __stdcall CreateCurve(const char *name)
{
	map<string, QUB_CurveFactory>::iterator ci = g_curveMakers.find(name);
	if ( ci != g_curveMakers.end() )
		return (ci->second)();
	else
		return CreateNullCurve();
}

extern "C" QUBFITS_API QTR_Impl* __stdcall ListCurves()
{
	QUB_Tree result = QUB_Tree::Create("Curves");
	for (map<string, QUB_CurveFactory>::iterator ci = g_curveMakers.begin();
		ci != g_curveMakers.end(); ++ci)
		result.appendChild("Curve").setData(ci->first);
	QTR_Impl *presult = result.getImpl();
	QTR_INCREF(presult);
	return presult;
}

extern "C" QUBFITS_API void __stdcall RegisterFitter(const char *name, QUB_FitterFactory factory)
{
	g_fitterMakers[name] = factory;
}

extern "C" QUBFITS_API IQUB_Fitter* __stdcall CreateFitter(const char *name)
{
	map<string, QUB_FitterFactory>::iterator fi = g_fitterMakers.find(name);
	if ( fi != g_fitterMakers.end() )
		return (fi->second)();
	else
		return CreateNullFitter();
}

extern "C" QUBFITS_API QTR_Impl* __stdcall ListFitters()
{
	QUB_Tree result = QUB_Tree::Create("Fitters");
	for (map<string, QUB_FitterFactory>::iterator fi = g_fitterMakers.begin();
		fi != g_fitterMakers.end(); ++fi)
		result.appendChild("Fitter").setData(fi->first);
	QTR_Impl *presult = result.getImpl();
	QTR_INCREF(presult);
	return presult;
}


double ChiSqr(double *yy, double *ff, double *ww, int n, int nfree)
{
    double x2 = 0.0;
	for (int i=0; i<n; ++i) {
        double var = yy[i] - ff[i];
        x2 += ww[i] * var * var;
	}
    return (x2 / nfree);
}

extern "C" QUBFITS_API double __stdcall ChiSqrUncertainty(IQUB_Curve* curve, QF_Data *data, double *ff, int ipar, double deltaPct/*0<d<1*/)
{
    int nfree = data->n - curve->getParamCount();
    double chiSqrFit = ChiSqr(data->yy, ff, data->ww, data->n, nfree);
    double chiSqrOff[2];
	double pval = curve->getParamVal(ipar);
	double delta = fabs(deltaPct * pval);
	if ( delta < 1.0e-8 )
		delta = 0.01;
	vector<double> ggv(data->n);
	double *gg = &(ggv[0]);
	for (int sign=-1, i=0; sign<2; sign+=2, ++i) { // -1, 1
		curve->setParamVal(ipar, pval + sign*delta, 0);
		curve->evalArrayAll(data, gg);
		chiSqrOff[i] = ChiSqr(data->yy, gg, data->ww, data->n, nfree);
	}
	curve->setParamVal(ipar, pval, 0);
	// avoiding zero-division and indicating sign problems:
	int sgn = 1;
	if ( nfree < 0 )
		sgn = -1;
	if ( nfree <= 0 )
		nfree = 1;
	double denom = (nfree * (chiSqrOff[0] + chiSqrOff[1] - 2.0*chiSqrFit));
	if ( denom <= 0.0 ) {
		sgn = -1;
		if ( denom == 0.0 )
			denom = 1.0;
		else
			denom = - denom;
	}
	double result = sgn * delta * sqrt(2.0 / denom);
#if defined(_WIN32)
	unsigned int fstat = _clearfp(); // just in case
	_fpreset();
#endif
	return result;
}



class QUB_NullCurve : public IQUB_Curve
{
public:
	QUB_NullCurve()
		: state( QUB_Tree::Create("Curve") ),
		  x0state( state["x0"] )
	{
		x0state.setNumData(QTR_TYPE_DOUBLE, 2, 1, (void*)0);
		x0state.dataAs(0,0,(double)0) = 0.0; // x0 = 0
		x0state.dataAs(1,0,(double)0) = 1.0; // const.
	}
	void __cdecl Free() { delete this; }
	IQUB_Evts * __cdecl Events() { return &events; }
	virtual void __cdecl setExtraVars(char **names, int count) {}
	const char * __cdecl getName() { return ""; }
	const char * __cdecl getEqn() { return ""; }
	void __cdecl setEqn(const char *x) {}
	int __cdecl canSetEqn() { return 0; }
	int __cdecl getComponentCount() { return 1; }
	void __cdecl setComponentCount(int x) {}
	int __cdecl getParamCount() { return 1; }
	const char * __cdecl getParamName(int ix) { return "x0"; }
	double __cdecl getParamVal(int ix) { return x0state.dataAs(0,0,(double)0); }
	void __cdecl setParamVal(int ix, double val, int sendEvent) {
		x0state.dataAs(0,0,(double)0) = val;
		if ( sendEvent )
			events.send("Curve ParamVal", this);
	}
	virtual int __cdecl getParamConst(int ix) { return (int) x0state.dataAs(1,0,(double)0); }
	virtual void __cdecl setParamConst(int ix, int isConst) {
		x0state.dataAs(1,0,(double)0) = (double) isConst;
		events.send("Curve ParamConst", this);
	}
	void __cdecl initParams(double *xx, double *yy, int n) {}
	void __cdecl evalArrayAll(QF_Data *data, double *yy) {
		for (int i=0; i<data->n; ++i)
			yy[i] = 0.0;
	}
	void __cdecl evalArrayOne(QF_Data *data, double *yy, int component) {
		for (int i=0; i<data->n; ++i)
			yy[i] = 0.0;
	}
	int __cdecl hasDeriv() { return 0; }
	void __cdecl evalDerivArray(QF_Data *data, double *dpar) {}
	const char * __cdecl getMessage() { return ""; }
	QTR_Impl* __cdecl getState() {
		QTR_Impl* pstate = state.getImpl();
		QTR_INCREF(pstate);
		return pstate;
	}
	void __cdecl updateState(QTR_Impl* pustate) {
		QUB_Tree ustate(pustate);
		QUB_Tree ux0 = * ustate.find("x0");
		if ( ux0.dataCount() > 0 )
			setParamVal(0, ux0.dataAs(0,0,(double)0), 1);
		if ( ux0.dataCount() > 1 )
			setParamConst(0, (int) ux0.dataAs(1,0,(double)0));
	}
private:
	QUB_NullCurve(const QUB_NullCurve& copy) {} // no copies
	QUB_NullCurve& operator= (const QUB_NullCurve& copy) { return *this; }
	QUB_Evts events;
	QUB_Tree state;
	QUB_Tree x0state;
};

QUBFITS_API IQUB_Curve* __stdcall CreateNullCurve()
{
	return new QUB_NullCurve;
}

class QUB_NullFitter : public IQUB_Fitter
{
public:
	QUB_NullFitter()
		: state( QUB_Tree::Create("Fitter") ) {}
	void __cdecl Free() { delete this; }
	IQUB_Evts * __cdecl Events() { return &events; }
	const char * __cdecl getName() { return ""; }
	int __cdecl getParamCount() { return 0; }
	const char * __cdecl getParamName(int ix) { return ""; }
	double __cdecl getParamVal(int ix) { return 0.0; }
	void __cdecl setParamVal(int ix, double val) {}
	int __cdecl Fit(IQUB_Curve* curve, QF_Data **data, int ndata) { return -1; }
	void __cdecl Stop() {} // from another thread
	int __cdecl getIterations() { return 0; }
	double __cdecl getParamErr(int ix) { return -1.0; }
	double * __cdecl getCurve() { return NULL; }
	const char * __cdecl getMessage() { return ""; } // if directed by event
	QTR_Impl* __cdecl getState() {
		QTR_Impl* pstate = state.getImpl();
		QTR_INCREF(pstate);
		return pstate;
	}
	void __cdecl updateState(QTR_Impl* state) {}
private:
	QUB_Evts events;
	QUB_Tree state;
};

QUBFITS_API IQUB_Fitter* __stdcall CreateNullFitter()
{
	return new QUB_NullFitter;
}

